#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the entrypoint for ingestion launching """
import json
import logging
import sys
import tarfile
from pathlib import Path

# pylint: disable=C0301, E0401, R0903, R1721
from typing import Tuple

import pendulum
from ingest_envoy.manifest_components import (
    INGESTION_ARTIFACTS_NAME,
    JSON,
    MANIFEST_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    IngestionManifestKey,
    InputGroup,
    InputScienceProduct,
    ManifestComponentIF,
    ManifestParameters,
    OutputGroup,
    OutputScienceProduct,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.std_img_manifest_utils import ImageIngestionProductsFinder
from ingest_envoy.utilities import (
    AncillaryProductType,
    IngestionManifestException,
    ScienceProductType,
    Telescope,
    find_output_tars,
)
from pendulum import DateTime

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

# pylint: disable=R0902, R0913


class IngestionManifest(ManifestComponentIF):
    """Encapsulates an ingestion manifest"""

    def __init__(
        self,
        telescope: Telescope,
        sp_type: ScienceProductType,
        staging_source_dir: Path,
        locator: str,
        # all except EVLA_EB and VLASS catalog manifest have input group
        input_group: InputGroup,
        # every manifest has at least one output group
        output_group: OutputGroup,
        # image manifest has this:
        additional_metadata: AbstractTextFile = None,
    ):
        self.staging_source_dir = staging_source_dir
        self.sp_type = sp_type
        self.locator = locator
        self.input_group = input_group
        self.output_group = output_group
        self.telescope = telescope

        self.parameters = self.build_ingest_parameters(additional_metadata)
        self.files_found = [file for file in self.staging_source_dir.iterdir()]

    def build_ingest_parameters(self, additional_metadata: AbstractTextFile):
        """
        Make the "parameters" section of the manifest

        :return:
        """
        if self.sp_type not in [ScienceProductType.EVLA_CAL, ScienceProductType.IMAGE]:
            raise NotImplementedError()

        if additional_metadata:
            params = ManifestParameters(
                telescope=self.telescope,
                reingest=False,
                ngas_ingest=True,
                calibrate=False,
                staging_source_dir=self.staging_source_dir,
                additional_metadata=additional_metadata,
            )

        elif self.sp_type == ScienceProductType.IMAGE:
            params = ManifestParameters(
                telescope=self.telescope,
                reingest=False,
                ngas_ingest=True,
                calibrate=False,
                staging_source_dir=self.staging_source_dir,
            )

        else:
            params = ManifestParameters(
                telescope=self.telescope,
                reingest=False,
                ngas_ingest=True,
                staging_source_dir=self.staging_source_dir,
            )

        return params

    def __eq__(self, other):
        if isinstance(other, IngestionManifest):
            return (
                other.parameters == self.parameters
                and other.input_group == self.input_group
                and other.output_group == self.output_group
            )

        return False

    def ingestion_path(self) -> Path:
        """
        Get the ingestion path associated with this manifest.

        :return:
        """
        return self.parameters.staging_source_dir

    def write(self) -> Path:
        """
        Write the manifest .json file.

        :return:
        """

        me_dict = self.to_json()
        output_path = self.staging_source_dir / MANIFEST_FILENAME

        to_write = json.dumps(me_dict, indent=4)
        with open(output_path, "w") as out:
            out.write(to_write)

        return output_path

    def to_json(self) -> JSON:
        """
        Turn this object into a JSON string suitable for writing to a file

        :return:
        """

        me_dict = dict(self.__dict__)

        to_return = {
            IngestionManifestKey.PARAMETERS.value: self.parameters.to_json(),
            IngestionManifestKey.INPUT_GROUP.value: me_dict[IngestionManifestKey.INPUT_GROUP.value].to_json(),
            IngestionManifestKey.OUTPUT_GROUP.value: me_dict[IngestionManifestKey.OUTPUT_GROUP.value].to_json(),
        }

        return to_return


class IngestionManifestBuilder:
    """Builds ingestion manifest and associated files from files in ingestion_path"""

    def __init__(
        self,
        staging_source_dir: Path,
        sp_type: str,
        locator: str,
        telescope: str,
        additional_metadata: AbstractTextFile = None,
    ):
        # get the telescope
        self.telescope = Telescope(telescope)
        if self.telescope is None:
            raise ValueError("telescope is required.")

        # get the directory containing the files to be ingested
        self.staging_source_dir = staging_source_dir

        # get the additional metadata, if any (EVLA CAL ingestions will have none)
        self.additional_metadata = additional_metadata

        # determine the science product type
        self.sp_type = ScienceProductType(sp_type)
        if self.sp_type is None or not isinstance(self.sp_type, ScienceProductType):
            raise ValueError(f"Invalid/unknown science product type: {sp_type}")
        if self.sp_type not in [ScienceProductType.EVLA_CAL, ScienceProductType.IMAGE]:
            raise NotImplementedError(f"Don't know yet how to build a {self.sp_type.value} manifest")

        self.locator = locator

        self.files_found = [file for file in staging_source_dir.iterdir()]
        if len(self.files_found) == 0:
            raise IngestionManifestException(f"No ingestion files found at {staging_source_dir}")

    def build(self) -> Tuple[IngestionManifest, Path]:
        """
        Using only -relevant- files in ingestion_path, write the manifest
        and produce other files required for ingestion.

        :return: the ingestion manifest and the file containing its JSON
        """

        if self.sp_type == ScienceProductType.EVLA_CAL:
            return self._build_evla_cal_manifest()

        return self._build_image_manifest()

    def _build_evla_cal_manifest(self):
        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            sp_type=self.sp_type,
            staging_source_dir=self.staging_source_dir,
            input_group=self._build_input_group(),
            output_group=self._build_evla_cal_output_group(),
        )

        # We can't create the ingestion artifacts tar quite yet,
        # because it will contain the yet-to-be-written manifest itself
        # (required for ingestion, evidently)
        artifacts_filename = self._build_artifacts_filename()
        artifacts_ap = AncillaryProduct(AncillaryProductType.INGESTION_ARTIFACTS, filename=artifacts_filename)

        if not manifest.output_group.ancillary_products:
            manifest.output_group.ancillary_products = []
        weblog_ap = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=WEBLOG_FILENAME)
        manifest.output_group.ancillary_products.append(weblog_ap)
        manifest.output_group.ancillary_products.append(artifacts_ap)

        manifest_file = manifest.write()
        artifacts_file = self.staging_source_dir / artifacts_filename
        self.write_ingestion_artifacts_tar(artifacts_file)

        return manifest, manifest_file

    def _build_image_manifest(self):
        """
        Image manifest has additional_metadata, and output group is way more complicated
        :return:
        """

        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            additional_metadata=self.additional_metadata,
            sp_type=self.sp_type,
            staging_source_dir=self.staging_source_dir,
            input_group=self._build_input_group(),
            output_group=self._build_imaging_output_group(),
        )

        artifacts_file = self.staging_source_dir / self._build_artifacts_filename()
        artifacts_ap = AncillaryProduct(type=AncillaryProductType.INGESTION_ARTIFACTS, filename=artifacts_file.name)
        manifest.output_group.ancillary_products.append(artifacts_ap)
        manifest_file = manifest.write()
        self.write_ingestion_artifacts_tar(artifacts_file)

        return manifest, manifest_file

    def _build_input_group(self):
        """
        Create the input group using the parameters.

        :return:
        """

        # N.B. this is sufficient for most types of ingestion,
        # but ALMA CALs will have multiple EB SPs, identified only by locator,
        # and VLBAs have no input group at all.
        sp_in = InputScienceProduct(locator=self.locator)

        return InputGroup([sp_in])

    def _build_evla_cal_output_group(self):
        """
        Create imaging manifest output group using the parameters
        and the contents of the staging dir.

        :return:
        """
        # find science product (we expect just one for this SP type)

        tars_found = find_output_tars(self.files_found, self.staging_source_dir)

        sci_prod = None
        for file in tars_found:
            sci_prod = OutputScienceProduct(type=self.sp_type, filename=file.name)
            break

        if sci_prod:
            return OutputGroup(science_products=[sci_prod])

    def _build_imaging_output_group(self) -> OutputGroup:
        """
        Create imaging manifest output group using the parameters
        and the contents of the staging dir.

        :return:
        """

        products_finder = ImageIngestionProductsFinder(self.staging_source_dir)
        science_products = products_finder.output_science_products
        ancillary_products = products_finder.ancillary_products

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    @staticmethod
    def _build_artifacts_filename() -> str:
        """
        Build unique manifest filename in standard format.

        :return: the filename
        """
        current_time = pendulum.now()
        timestamp = format_timestamp(current_time)
        return f"{INGESTION_ARTIFACTS_NAME}{timestamp}{TARFILE_EXT}"

    def write_ingestion_artifacts_tar(self, artifacts_path: Path) -> tarfile.TarFile:
        """
        Take the list of files and build a tar for inclusion into the archive.
        This happens in the staging area for ingestion.
        The EVLA CAL tar will contain just the manifest.

        :return: a .tar archive of the ingestion artifacts
        """

        addl_md_file = None
        if self.additional_metadata:
            # find the additional metadata
            addl_md_filename = self.additional_metadata.filename
            addl_md_file = self.staging_source_dir / addl_md_filename

        with tarfile.open(artifacts_path, "w") as ingestion_artifacts_tar:
            if addl_md_file:
                ingestion_artifacts_tar.add(addl_md_file)

            # The manifest file itself is considered an ingestion artifact.
            # (It's turtles all the way down.)
            manifest_file = self.staging_source_dir / MANIFEST_FILENAME
            ingestion_artifacts_tar.add(manifest_file)

        return ingestion_artifacts_tar


def format_timestamp(datetime: DateTime) -> str:
    """
    Format the current time as
        2021_07_01T13_49_17.237

    :param datetime: current timestamp
    :return: timestamp suitable for ingestion manifest filename
    """

    return datetime.format("YYYY_MM_DDThh_mm_ss.SSS")


def find_manifest(ingestion_path: Path) -> Path:
    """
    Find the ingestion manifest at this ingestion path.

    :param ingestion_path: home of ingestion files
    :return:
    """
    for json_file in ingestion_path.glob(MANIFEST_FILENAME):
        return json_file

    raise FileNotFoundError(f"No ingestion manifest found at {ingestion_path}")
